@@ -1,5 +1,5 @@
 /*
- * Copyright 2013 MongoDB, Inc.
+ * Copyright 2009-present MongoDB, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -15,20 +15,22 @@
  */
 
 
-#include "mongoc-aggregate-private.h"
-#include "mongoc-client-private.h"
-#include "mongoc-collection.h"
-#include "mongoc-collection-private.h"
-#include "mongoc-cursor.h"
-#include "mongoc-cursor-private.h"
-#include "mongoc-database.h"
-#include "mongoc-database-private.h"
-#include "mongoc-error.h"
-#include "mongoc-log.h"
-#include "mongoc-trace-private.h"
-#include "mongoc-util-private.h"
-#include "mongoc-write-concern-private.h"
-#include "mongoc-change-stream-private.h"
+#include <mongoc/mongoc-aggregate-private.h>
+#include <mongoc/mongoc-client-private.h>
+#include <mongoc/mongoc-collection.h>
+#include <mongoc/mongoc-collection-private.h>
+#include <mongoc/mongoc-cursor.h>
+#include <mongoc/mongoc-cursor-private.h>
+#include <mongoc/mongoc-database.h>
+#include <mongoc/mongoc-database-private.h>
+#include <mongoc/mongoc-error.h>
+#include <mongoc/mongoc-log.h>
+#include <mongoc/mongoc-trace-private.h>
+#include <mongoc/mongoc-util-private.h>
+#include <mongoc/mongoc-write-concern-private.h>
+#include <mongoc/mongoc-change-stream-private.h>
+
+#include <common-bson-dsl-private.h>
 
 #undef MONGOC_LOG_DOMAIN
 #define MONGOC_LOG_DOMAIN "database"
@@ -70,12 +72,9 @@ _mongoc_database_new (mongoc_client_t *client,
 
    db = (mongoc_database_t *) bson_malloc0 (sizeof *db);
    db->client = client;
-   db->write_concern = write_concern ? mongoc_write_concern_copy (write_concern)
-                                     : mongoc_write_concern_new ();
-   db->read_concern = read_concern ? mongoc_read_concern_copy (read_concern)
-                                   : mongoc_read_concern_new ();
-   db->read_prefs = read_prefs ? mongoc_read_prefs_copy (read_prefs)
-                               : mongoc_read_prefs_new (MONGOC_READ_PRIMARY);
+   db->write_concern = write_concern ? mongoc_write_concern_copy (write_concern) : mongoc_write_concern_new ();
+   db->read_concern = read_concern ? mongoc_read_concern_copy (read_concern) : mongoc_read_concern_new ();
+   db->read_prefs = read_prefs ? mongoc_read_prefs_copy (read_prefs) : mongoc_read_prefs_new (MONGOC_READ_PRIMARY);
 
    db->name = bson_strdup (name);
 
@@ -172,11 +171,8 @@ mongoc_database_copy (mongoc_database_t *database)
 
    BSON_ASSERT_PARAM (database);
 
-   RETURN (_mongoc_database_new (database->client,
-                                 database->name,
-                                 database->read_prefs,
-                                 database->read_concern,
-                                 database->write_concern));
+   RETURN (_mongoc_database_new (
+      database->client, database->name, database->read_prefs, database->read_concern, database->write_concern));
 }
 
 mongoc_cursor_t *
@@ -211,8 +207,7 @@ mongoc_database_command (mongoc_database_t *database,
     */
 
    /* flags, skip, limit, batch_size, fields are unused */
-   cursor = _mongoc_cursor_cmd_deprecated_new (
-      database->client, ns, command, read_prefs);
+   cursor = _mongoc_cursor_cmd_deprecated_new (database->client, ns, command, read_prefs);
    bson_free (ns);
    return cursor;
 }
@@ -274,11 +269,8 @@ mongoc_database_read_command_with_opts (mongoc_database_t *database,
 
 
 bool
-mongoc_database_write_command_with_opts (mongoc_database_t *database,
-                                         const bson_t *command,
-                                         const bson_t *opts,
-                                         bson_t *reply,
-                                         bson_error_t *error)
+mongoc_database_write_command_with_opts (
+   mongoc_database_t *database, const bson_t *command, const bson_t *opts, bson_t *reply, bson_error_t *error)
 {
    return _mongoc_client_command_with_opts (database->client,
                                             database->name,
@@ -296,13 +288,12 @@ mongoc_database_write_command_with_opts (mongoc_database_t *database,
 
 
 bool
-mongoc_database_read_write_command_with_opts (
-   mongoc_database_t *database,
-   const bson_t *command,
-   const mongoc_read_prefs_t *read_prefs /* IGNORED */,
-   const bson_t *opts,
-   bson_t *reply,
-   bson_error_t *error)
+mongoc_database_read_write_command_with_opts (mongoc_database_t *database,
+                                              const bson_t *command,
+                                              const mongoc_read_prefs_t *read_prefs /* IGNORED */,
+                                              const bson_t *opts,
+                                              bson_t *reply,
+                                              bson_error_t *error)
 {
    return _mongoc_client_command_with_opts (database->client,
                                             database->name,
@@ -369,9 +360,7 @@ mongoc_database_drop (mongoc_database_t *database, bson_error_t *error)
 
 
 bool
-mongoc_database_drop_with_opts (mongoc_database_t *database,
-                                const bson_t *opts,
-                                bson_error_t *error)
+mongoc_database_drop_with_opts (mongoc_database_t *database, const bson_t *opts, bson_error_t *error)
 {
    bool ret;
    bson_t cmd;
@@ -400,9 +389,7 @@ mongoc_database_drop_with_opts (mongoc_database_t *database,
 
 
 bool
-mongoc_database_remove_user (mongoc_database_t *database,
-                             const char *username,
-                             bson_error_t *error)
+mongoc_database_remove_user (mongoc_database_t *database, const char *username, bson_error_t *error)
 {
    bson_t cmd;
    bool ret;
@@ -422,8 +409,7 @@ mongoc_database_remove_user (mongoc_database_t *database,
 
 
 bool
-mongoc_database_remove_all_users (mongoc_database_t *database,
-                                  bson_error_t *error)
+mongoc_database_remove_all_users (mongoc_database_t *database, bson_error_t *error)
 {
    bson_t cmd;
    bool ret;
@@ -464,7 +450,7 @@ mongoc_database_add_user (mongoc_database_t *database,
                           bson_error_t *error)
 {
    bson_t cmd;
-   bson_t ar;
+   bson_array_builder_t *ar;
    bool ret = false;
 
    ENTRY;
@@ -482,8 +468,8 @@ mongoc_database_add_user (mongoc_database_t *database,
    if (roles) {
       BSON_APPEND_ARRAY (&cmd, "roles", roles);
    } else {
-      bson_append_array_begin (&cmd, "roles", 5, &ar);
-      bson_append_array_end (&cmd, &ar);
+      bson_append_array_builder_begin (&cmd, "roles", 5, &ar);
+      bson_append_array_builder_end (&cmd, ar);
    }
 
    ret = mongoc_database_command_simple (database, &cmd, NULL, NULL, error);
@@ -535,8 +521,7 @@ mongoc_database_get_read_prefs (const mongoc_database_t *database) /* IN */
  */
 
 void
-mongoc_database_set_read_prefs (mongoc_database_t *database,
-                                const mongoc_read_prefs_t *read_prefs)
+mongoc_database_set_read_prefs (mongoc_database_t *database, const mongoc_read_prefs_t *read_prefs)
 {
    BSON_ASSERT_PARAM (database);
 
@@ -593,8 +578,7 @@ mongoc_database_get_read_concern (const mongoc_database_t *database)
  */
 
 void
-mongoc_database_set_read_concern (mongoc_database_t *database,
-                                  const mongoc_read_concern_t *read_concern)
+mongoc_database_set_read_concern (mongoc_database_t *database, const mongoc_read_concern_t *read_concern)
 {
    BSON_ASSERT_PARAM (database);
 
@@ -651,8 +635,7 @@ mongoc_database_get_write_concern (const mongoc_database_t *database)
  */
 
 void
-mongoc_database_set_write_concern (mongoc_database_t *database,
-                                   const mongoc_write_concern_t *write_concern)
+mongoc_database_set_write_concern (mongoc_database_t *database, const mongoc_write_concern_t *write_concern)
 {
    BSON_ASSERT_PARAM (database);
 
@@ -685,9 +668,7 @@ mongoc_database_set_write_concern (mongoc_database_t *database,
  * Returns: %true if @name exists, otherwise %false. @error may be set.
  */
 bool
-mongoc_database_has_collection (mongoc_database_t *database,
-                                const char *name,
-                                bson_error_t *error)
+mongoc_database_has_collection (mongoc_database_t *database, const char *name, bson_error_t *error)
 {
    bson_iter_t col_iter;
    bool ret = false;
@@ -712,9 +693,7 @@ mongoc_database_has_collection (mongoc_database_t *database,
 
    cursor = mongoc_database_find_collections_with_opts (database, &opts);
    while (mongoc_cursor_next (cursor, &doc)) {
-      if (bson_iter_init (&col_iter, doc) &&
-          bson_iter_find (&col_iter, "name") &&
-          BSON_ITER_HOLDS_UTF8 (&col_iter) &&
+      if (bson_iter_init (&col_iter, doc) && bson_iter_find (&col_iter, "name") && BSON_ITER_HOLDS_UTF8 (&col_iter) &&
           (cur_name = bson_iter_utf8 (&col_iter, NULL))) {
          if (!strcmp (cur_name, name)) {
             ret = true;
@@ -734,9 +713,7 @@ cleanup:
 
 
 mongoc_cursor_t *
-mongoc_database_find_collections (mongoc_database_t *database,
-                                  const bson_t *filter,
-                                  bson_error_t *error)
+mongoc_database_find_collections (mongoc_database_t *database, const bson_t *filter, bson_error_t *error)
 {
    bson_t opts = BSON_INITIALIZER;
    mongoc_cursor_t *cursor;
@@ -745,10 +722,7 @@ mongoc_database_find_collections (mongoc_database_t *database,
 
    if (filter) {
       if (!BSON_APPEND_DOCUMENT (&opts, "filter", filter)) {
-         bson_set_error (error,
-                         MONGOC_ERROR_BSON,
-                         MONGOC_ERROR_BSON_INVALID,
-                         "Invalid 'filter' parameter.");
+         bson_set_error (error, MONGOC_ERROR_BSON, MONGOC_ERROR_BSON_INVALID, "Invalid 'filter' parameter.");
          bson_destroy (&opts);
          return NULL;
       }
@@ -769,8 +743,7 @@ mongoc_database_find_collections (mongoc_database_t *database,
 
 
 mongoc_cursor_t *
-mongoc_database_find_collections_with_opts (mongoc_database_t *database,
-                                            const bson_t *opts)
+mongoc_database_find_collections_with_opts (mongoc_database_t *database, const bson_t *opts)
 {
    mongoc_cursor_t *cursor;
    bson_t cmd = BSON_INITIALIZER;
@@ -781,8 +754,7 @@ mongoc_database_find_collections_with_opts (mongoc_database_t *database,
 
    /* Enumerate Collections Spec: "run listCollections on the primary node in
     * replicaset mode" */
-   cursor = _mongoc_cursor_cmd_new (
-      database->client, database->name, &cmd, opts, NULL, NULL, NULL);
+   cursor = _mongoc_cursor_cmd_new (database->client, database->name, &cmd, opts, NULL, NULL, NULL);
    if (cursor->error.domain == 0) {
       _mongoc_cursor_prime (cursor);
    }
@@ -793,18 +765,14 @@ mongoc_database_find_collections_with_opts (mongoc_database_t *database,
 
 
 char **
-mongoc_database_get_collection_names (mongoc_database_t *database,
-                                      bson_error_t *error)
+mongoc_database_get_collection_names (mongoc_database_t *database, bson_error_t *error)
 {
-   return mongoc_database_get_collection_names_with_opts (
-      database, NULL, error);
+   return mongoc_database_get_collection_names_with_opts (database, NULL, error);
 }
 
 
 char **
-mongoc_database_get_collection_names_with_opts (mongoc_database_t *database,
-                                                const bson_t *opts,
-                                                bson_error_t *error)
+mongoc_database_get_collection_names_with_opts (mongoc_database_t *database, const bson_t *opts, bson_error_t *error)
 {
    bson_t opts_copy;
    bson_iter_t col;
@@ -834,8 +802,8 @@ mongoc_database_get_collection_names_with_opts (mongoc_database_t *database,
    _mongoc_array_init (&strv_buf, sizeof (char *));
 
    while (mongoc_cursor_next (cursor, &doc)) {
-      if (bson_iter_init (&col, doc) && bson_iter_find (&col, "name") &&
-          BSON_ITER_HOLDS_UTF8 (&col) && (name = bson_iter_utf8 (&col, NULL))) {
+      if (bson_iter_init (&col, doc) && bson_iter_find (&col, "name") && BSON_ITER_HOLDS_UTF8 (&col) &&
+          (name = bson_iter_utf8 (&col, NULL))) {
          namecopy = bson_strdup (name);
          _mongoc_array_append_val (&strv_buf, namecopy);
       }
@@ -860,10 +828,7 @@ mongoc_database_get_collection_names_with_opts (mongoc_database_t *database,
 }
 
 static mongoc_collection_t *
-create_collection (mongoc_database_t *database,
-                   const char *name,
-                   const bson_t *opts,
-                   bson_error_t *error)
+create_collection (mongoc_database_t *database, const char *name, const bson_t *opts, bson_error_t *error)
 {
    mongoc_collection_t *collection = NULL;
    bson_iter_t iter;
@@ -874,11 +839,8 @@ create_collection (mongoc_database_t *database,
    BSON_ASSERT_PARAM (name);
 
    if (strchr (name, '$')) {
-      bson_set_error (error,
-                      MONGOC_ERROR_NAMESPACE,
-                      MONGOC_ERROR_NAMESPACE_INVALID,
-                      "The namespace \"%s\" is invalid.",
-                      name);
+      bson_set_error (
+         error, MONGOC_ERROR_NAMESPACE, MONGOC_ERROR_NAMESPACE_INVALID, "The namespace \"%s\" is invalid.", name);
       return NULL;
    }
 
@@ -903,11 +865,10 @@ create_collection (mongoc_database_t *database,
             return NULL;
          }
          if (!capped) {
-            bson_set_error (
-               error,
-               MONGOC_ERROR_COMMAND,
-               MONGOC_ERROR_COMMAND_INVALID_ARG,
-               "The \"size\" parameter requires {\"capped\": true}");
+            bson_set_error (error,
+                            MONGOC_ERROR_COMMAND,
+                            MONGOC_ERROR_COMMAND_INVALID_ARG,
+                            "The \"size\" parameter requires {\"capped\": true}");
             return NULL;
          }
       }
@@ -921,22 +882,20 @@ create_collection (mongoc_database_t *database,
             return NULL;
          }
          if (!capped) {
-            bson_set_error (
-               error,
-               MONGOC_ERROR_COMMAND,
-               MONGOC_ERROR_COMMAND_INVALID_ARG,
-               "The \"max\" parameter requires {\"capped\": true}");
+            bson_set_error (error,
+                            MONGOC_ERROR_COMMAND,
+                            MONGOC_ERROR_COMMAND_INVALID_ARG,
+                            "The \"max\" parameter requires {\"capped\": true}");
             return NULL;
          }
       }
 
       if (bson_iter_init_find (&iter, opts, "storageEngine")) {
          if (!BSON_ITER_HOLDS_DOCUMENT (&iter)) {
-            bson_set_error (
-               error,
-               MONGOC_ERROR_COMMAND,
-               MONGOC_ERROR_COMMAND_INVALID_ARG,
-               "The \"storageEngine\" parameter must be a document");
+            bson_set_error (error,
+                            MONGOC_ERROR_COMMAND,
+                            MONGOC_ERROR_COMMAND_INVALID_ARG,
+                            "The \"storageEngine\" parameter must be a document");
 
             return NULL;
          }
@@ -953,11 +912,10 @@ create_collection (mongoc_database_t *database,
 
             if (bson_iter_find (&iter, "configString")) {
                if (!BSON_ITER_HOLDS_UTF8 (&iter)) {
-                  bson_set_error (
-                     error,
-                     MONGOC_ERROR_COMMAND,
-                     MONGOC_ERROR_COMMAND_INVALID_ARG,
-                     "The \"configString\" parameter must be a string");
+                  bson_set_error (error,
+                                  MONGOC_ERROR_COMMAND,
+                                  MONGOC_ERROR_COMMAND_INVALID_ARG,
+                                  "The \"configString\" parameter must be a string");
                   return NULL;
                }
             } else {
@@ -988,12 +946,8 @@ create_collection (mongoc_database_t *database,
                                          database->write_concern,
                                          NULL, /* reply */
                                          error)) {
-      collection = _mongoc_collection_new (database->client,
-                                           database->name,
-                                           name,
-                                           database->read_prefs,
-                                           database->read_concern,
-                                           database->write_concern);
+      collection = _mongoc_collection_new (
+         database->client, database->name, name, database->read_prefs, database->read_concern, database->write_concern);
    }
 
    bson_destroy (&cmd);
@@ -1002,26 +956,23 @@ create_collection (mongoc_database_t *database,
 }
 
 char *
-_mongoc_get_encryptedField_state_collection (
-   const bson_t *encryptedFields,
-   const char *data_collection,
-   const char *state_collection_suffix,
-   bson_error_t *error)
+_mongoc_get_encryptedField_state_collection (const bson_t *encryptedFields,
+                                             const char *data_collection,
+                                             const char *state_collection_suffix,
+                                             bson_error_t *error)
 {
    bson_iter_t iter;
    const char *fieldName = NULL;
 
    if (0 == strcmp (state_collection_suffix, "esc")) {
       fieldName = "escCollection";
-   } else if (0 == strcmp (state_collection_suffix, "ecc")) {
-      fieldName = "eccCollection";
    } else if (0 == strcmp (state_collection_suffix, "ecoc")) {
       fieldName = "ecocCollection";
    } else {
       bson_set_error (error,
                       MONGOC_ERROR_COMMAND,
                       MONGOC_ERROR_COMMAND_INVALID_ARG,
-                      "expected state_collection_suffix to be 'esc', 'ecc', or "
+                      "expected state_collection_suffix to be 'esc' or "
                       "'ecoc', got: %s",
                       state_collection_suffix);
       return NULL;
@@ -1038,8 +989,7 @@ _mongoc_get_encryptedField_state_collection (
       }
       return bson_strdup (bson_iter_utf8 (&iter, NULL));
    }
-   return bson_strdup_printf (
-      "enxcol_.%s.%s", data_collection, state_collection_suffix);
+   return bson_strdup_printf ("enxcol_.%s.%s", data_collection, state_collection_suffix);
 }
 
 static bool
@@ -1054,23 +1004,13 @@ create_encField_state_collection (mongoc_database_t *database,
    bool ok = false;
    bson_t opts = BSON_INITIALIZER;
 
-   state_collection = _mongoc_get_encryptedField_state_collection (
-      encryptedFields, data_collection, state_collection_suffix, error);
+   state_collection =
+      _mongoc_get_encryptedField_state_collection (encryptedFields, data_collection, state_collection_suffix, error);
    if (!state_collection) {
       goto fail;
    }
 
-   BCON_APPEND (&opts,
-                "clusteredIndex",
-                "{",
-                "key",
-                "{",
-                "_id",
-                BCON_INT32 (1),
-                "}",
-                "unique",
-                BCON_BOOL (true),
-                "}");
+   BCON_APPEND (&opts, "clusteredIndex", "{", "key", "{", "_id", BCON_INT32 (1), "}", "unique", BCON_BOOL (true), "}");
 
    collection = create_collection (database, state_collection, &opts, error);
    if (collection == NULL) {
@@ -1095,13 +1035,36 @@ create_collection_with_encryptedFields (mongoc_database_t *database,
    mongoc_collection_t *dataCollection = NULL;
    bool ok = false;
    bson_t *cc_opts = NULL;
-   bool state_collections_ok =
-      create_encField_state_collection (
-         database, encryptedFields, name, "esc", error) &&
-      create_encField_state_collection (
-         database, encryptedFields, name, "ecc", error) &&
-      create_encField_state_collection (
-         database, encryptedFields, name, "ecoc", error);
+
+   // Check the wire version to ensure server is 7.0.0 or newer.
+   {
+      const mongoc_ss_log_context_t ss_log_context = {.operation = "createCollection"};
+      mongoc_server_stream_t *stream = mongoc_cluster_stream_for_writes (&database->client->cluster,
+                                                                         &ss_log_context,
+                                                                         NULL /* client session */,
+                                                                         NULL /* deprioritized servers */,
+                                                                         NULL /* reply */,
+                                                                         error);
+      if (!stream) {
+         goto fail;
+      }
+      if (stream->sd->max_wire_version < WIRE_VERSION_7_0) {
+         bson_set_error (error,
+                         MONGOC_ERROR_PROTOCOL,
+                         MONGOC_ERROR_PROTOCOL_BAD_WIRE_VERSION,
+                         "Driver support of Queryable Encryption is incompatible "
+                         "with server. Upgrade server to use Queryable Encryption. "
+                         "Got maxWireVersion %" PRId32 " but need maxWireVersion >= %d",
+                         stream->sd->max_wire_version,
+                         WIRE_VERSION_7_0);
+         mongoc_server_stream_cleanup (stream);
+         goto fail;
+      }
+      mongoc_server_stream_cleanup (stream);
+   }
+
+   bool state_collections_ok = create_encField_state_collection (database, encryptedFields, name, "esc", error) &&
+                               create_encField_state_collection (database, encryptedFields, name, "ecoc", error);
    if (!state_collections_ok) {
       // Failed to create one or more state collections
       goto fail;
@@ -1110,10 +1073,8 @@ create_collection_with_encryptedFields (mongoc_database_t *database,
    /* Create data collection. */
    cc_opts = bson_copy (opts);
    if (!BSON_APPEND_DOCUMENT (cc_opts, "encryptedFields", encryptedFields)) {
-      bson_set_error (error,
-                      MONGOC_ERROR_COMMAND,
-                      MONGOC_ERROR_COMMAND_INVALID_ARG,
-                      "unable to append encryptedFields");
+      bson_set_error (
+         error, MONGOC_ERROR_COMMAND, MONGOC_ERROR_COMMAND_INVALID_ARG, "unable to append encryptedFields");
       goto fail;
    }
    dataCollection = create_collection (database, name, cc_opts, error);
@@ -1124,26 +1085,10 @@ create_collection_with_encryptedFields (mongoc_database_t *database,
    /* Create index on __safeContent__. */
    {
       bson_t *keys = BCON_NEW ("__safeContent__", BCON_INT32 (1));
-      char *index_name;
-      bson_t *create_indexes;
-
-      index_name = mongoc_collection_keys_to_index_string (keys);
-      create_indexes = BCON_NEW ("createIndexes",
-                                 BCON_UTF8 (name),
-                                 "indexes",
-                                 "[",
-                                 "{",
-                                 "key",
-                                 BCON_DOCUMENT (keys),
-                                 "name",
-                                 BCON_UTF8 (index_name),
-                                 "}",
-                                 "]");
-
-      ok = mongoc_database_write_command_with_opts (
-         database, create_indexes, NULL /* opts */, NULL /* reply */, error);
-      bson_destroy (create_indexes);
-      bson_free (index_name);
+      mongoc_index_model_t *im = mongoc_index_model_new (keys, NULL /* opts */);
+      ok =
+         mongoc_collection_create_indexes_with_opts (dataCollection, &im, 1, NULL /* opts */, NULL /* reply */, error);
+      mongoc_index_model_destroy (im);
       bson_destroy (keys);
       if (!ok) {
          goto fail;
@@ -1162,12 +1107,11 @@ fail:
 }
 
 bool
-_mongoc_get_encryptedFields_from_map (mongoc_client_t *client,
-                                      const char *dbName,
-                                      const char *collName,
-                                      bson_t *encryptedFields,
-                                      bson_error_t *error)
+_mongoc_get_encryptedFields_from_map (
+   mongoc_client_t *client, const char *dbName, const char *collName, bson_t *encryptedFields, bson_error_t *error)
 {
+   BSON_ASSERT_PARAM (client);
+
    const bson_t *efMap = client->topology->encrypted_fields_map;
 
    bson_init (encryptedFields);
@@ -1195,12 +1139,11 @@ _mongoc_get_encryptedFields_from_map (mongoc_client_t *client,
 }
 
 bool
-_mongoc_get_encryptedFields_from_server (mongoc_client_t *client,
-                                         const char *dbName,
-                                         const char *collName,
-                                         bson_t *encryptedFields,
-                                         bson_error_t *error)
+_mongoc_get_encryptedFields_from_server (
+   mongoc_client_t *client, const char *dbName, const char *collName, bson_t *encryptedFields, bson_error_t *error)
 {
+   BSON_ASSERT_PARAM (client);
+
    mongoc_database_t *db = mongoc_client_get_database (client, dbName);
    bson_t *opts = BCON_NEW ("filter", "{", "name", BCON_UTF8 (collName), "}");
    mongoc_cursor_t *cursor;
@@ -1218,10 +1161,8 @@ _mongoc_get_encryptedFields_from_server (mongoc_client_t *client,
       /* Check if the collInfo has options.encryptedFields. */
       bson_iter_t iter;
       if (!bson_iter_init (&iter, collInfo)) {
-         bson_set_error (error,
-                         MONGOC_ERROR_COMMAND,
-                         MONGOC_ERROR_COMMAND_INVALID_ARG,
-                         "unable to iterate listCollections result");
+         bson_set_error (
+            error, MONGOC_ERROR_COMMAND, MONGOC_ERROR_COMMAND_INVALID_ARG, "unable to iterate listCollections result");
          goto fail;
       }
 
@@ -1246,59 +1187,104 @@ fail:
    return ret;
 }
 
+bool
+_mongoc_get_collection_encryptedFields (mongoc_client_t *client,
+                                        const char *dbName,
+                                        const char *collName,
+                                        const bson_t *opts,
+                                        bool checkEncryptedFieldsMap,
+                                        bson_t *encryptedFields,
+                                        bson_error_t *error)
+{
+   BSON_ASSERT_PARAM (client);
+   BSON_ASSERT_PARAM (dbName);
+   BSON_ASSERT_PARAM (collName);
+   BSON_OPTIONAL_PARAM (opts);
+   BSON_ASSERT_PARAM (encryptedFields);
+   BSON_OPTIONAL_PARAM (error);
+
+   bson_init (encryptedFields); // Initially empty
+
+   if (opts) {
+      // We have collection options, which may have encryptedFields in it
+      bool found = false;
+      bsonParse (*opts,
+                 find (key ("encryptedFields"),
+                       if (not(type (doc)), then (error ("'encryptedFields' should be a document"))),
+                       // Update encryptedFields to be a reference to the subdocument:
+                       storeDocRef (*encryptedFields),
+                       do (found = true)));
+      if (bsonParseError) {
+         // Error while parsing
+         bson_set_error (error,
+                         MONGOC_ERROR_COMMAND,
+                         MONGOC_ERROR_COMMAND_INVALID_ARG,
+                         "Invalid createCollection command options: %s",
+                         bsonParseError);
+         return false;
+      } else if (found) {
+         // Found it!
+         return true;
+      } else {
+         // Nothing found in the options
+      }
+   }
+
+   // Look in the encryptedFieldsMap based on this collection name
+   if (checkEncryptedFieldsMap &&
+       !_mongoc_get_encryptedFields_from_map (client, dbName, collName, encryptedFields, error)) {
+      // Error during lookup.
+      return false;
+   }
+
+   // No error. We may or may not have found encryptedFields. The caller
+   // determines this by checking if encryptedFields is empty.
+   return true;
+}
+
 mongoc_collection_t *
 mongoc_database_create_collection (mongoc_database_t *database,
                                    const char *name,
                                    const bson_t *opts,
                                    bson_error_t *error)
 {
-   bson_iter_t iter;
+   BSON_ASSERT_PARAM (database);
+   BSON_ASSERT_PARAM (name);
+   BSON_OPTIONAL_PARAM (opts);
+   BSON_OPTIONAL_PARAM (error);
+
    bson_t encryptedFields = BSON_INITIALIZER;
-
-   if (opts && bson_iter_init_find (&iter, opts, "encryptedFields")) {
-      if (!_mongoc_iter_document_as_bson (&iter, &encryptedFields, error)) {
-         return NULL;
-      }
-   }
-
-   if (bson_empty (&encryptedFields)) {
-      if (!_mongoc_get_encryptedFields_from_map (
-             database->client,
-             mongoc_database_get_name (database),
-             name,
-             &encryptedFields,
-             error)) {
-         return NULL;
-      }
+   if (!_mongoc_get_collection_encryptedFields (database->client,
+                                                mongoc_database_get_name (database),
+                                                name,
+                                                opts,
+                                                true /* checkEncryptedFieldsMap */,
+                                                &encryptedFields,
+                                                error)) {
+      // Error during fields lookup
+      bson_destroy (&encryptedFields);
+      return NULL;
    }
 
    if (!bson_empty (&encryptedFields)) {
-      bson_t opts_without_encryptedFields = BSON_INITIALIZER;
+      // Clone 'opts' without the encryptedFields element
+      bsonBuildDecl (opts_without_encryptedFields, if (opts, then (insert (*opts, not(key ("encryptedFields"))))));
 
-      if (opts) {
-         bson_copy_to_excluding_noinit (
-            opts, &opts_without_encryptedFields, "encryptedFields", NULL);
-      }
+      mongoc_collection_t *ret = create_collection_with_encryptedFields (
+         database, name, &opts_without_encryptedFields, &encryptedFields, error);
 
-      mongoc_collection_t *ret =
-         create_collection_with_encryptedFields (database,
-                                                 name,
-                                                 &opts_without_encryptedFields,
-                                                 &encryptedFields,
-                                                 error);
-
-      bson_destroy (&opts_without_encryptedFields);
       bson_destroy (&encryptedFields);
+      bson_destroy (&opts_without_encryptedFields);
       return ret;
    }
 
+   bson_destroy (&encryptedFields);
    return create_collection (database, name, opts, error);
 }
 
 
 mongoc_collection_t *
-mongoc_database_get_collection (mongoc_database_t *database,
-                                const char *collection)
+mongoc_database_get_collection (mongoc_database_t *database, const char *collection)
 {
    BSON_ASSERT_PARAM (database);
    BSON_ASSERT_PARAM (collection);
@@ -1322,9 +1308,7 @@ mongoc_database_get_name (mongoc_database_t *database)
 
 
 mongoc_change_stream_t *
-mongoc_database_watch (const mongoc_database_t *db,
-                       const bson_t *pipeline,
-                       const bson_t *opts)
+mongoc_database_watch (const mongoc_database_t *db, const bson_t *pipeline, const bson_t *opts)
 {
    return _mongoc_change_stream_new_from_database (db, pipeline, opts);
 }
